/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.net.*;
import lispwrapper.*;
import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import img.fact.cexpr.*;
import img.fact.*;
import img.fact.xml.*;
import img.util.*;

/** I'm a little client application that talks to fact using the corba
 * interface.  
 * This class contains the basic functionality that an applet or 
 * application should provide. Classes can inherit from this and give
 * extra stuff.
 * @author SKB */

public abstract class BasicClient extends JApplet
implements ActionListener, CompilerLogger {


  // Thing which will do some conversion for you.
private XMLToLisp xmlConverter;

  /** Build a new client. Launches an application that can then be
      used to connect to a server */
public void init() {

  // This is rather horrid. The parsers are static, i.e. there's one
  // that's shared -- so we'd better make sure that things don't get
  // trashed if multiple clients are querying the kb. I think we'll
  // need to rethink this at a later date. 
  new img.fact.cexpr.Parser(new java.io.StringReader(""));


  // setTitle("Client");

  Container content = this.getContentPane();
  content.setLayout(new BorderLayout(20,20));

  // Some fields that indicate the status of the client.
  JPanel p = new JPanel(new FlowLayout(FlowLayout.LEFT));
  conn = new JTextField("", 7);
  conn.setBorder(new EmptyBorder(new Insets(0,0,0,0)));
  conn.setHorizontalAlignment(JTextField.CENTER);
  conn.setEditable(false);

  trans = new JTextField ("", 7);
  trans.setBorder(new EmptyBorder(new Insets(0,0,0,0)));
  trans.setHorizontalAlignment(JTextField.CENTER);
  trans.setEditable(false);

  comp = new JTextField ("", 7);
  comp.setBorder(new EmptyBorder(new Insets(0,0,0,0)));
  comp.setHorizontalAlignment(JTextField.CENTER);
  comp.setEditable(false);
  comp.setBackground(null);
  comp.setForeground(Color.black);

  serv = new JTextField ("", 7);
  serv.setBorder(new EmptyBorder(new Insets(0,0,0,0)));
  serv.setHorizontalAlignment(JTextField.CENTER);
  serv.setEditable(false);
  serv.setBackground(null);
  serv.setForeground(Color.black);

  p.add(new JLabel("Status: "));
  p.add(conn);
  p.add(trans);
  p.add(comp);
  p.add(serv);
  content.add(p, "North");

  p = new JPanel(new BorderLayout(5,5));
  tells = new JTextArea();
  tells.setFont(new Font("SanSerif", Font.PLAIN, 8));
  p.setBorder(new TitledBorder("Transaction Tells"));
  p.add(new JScrollPane(tells));
  content.add(p, "Center");


  JPanel resultAndLog = new JPanel(new GridLayout(2,1));
  p = new JPanel(new BorderLayout(5,5));
  result = new JTextArea();
  result.setEditable(false);
  result.setLineWrap(true);
  result.setWrapStyleWord(true);
  resultScroll = new JScrollPane(result);
  p.setBorder(new TitledBorder("Result"));
  p.add(resultScroll);
  resultAndLog.add(p);

  p = new JPanel(new BorderLayout(5,5));
  report = new JTextArea(4,30);
  report.setEditable(false);
  p.setBorder(new TitledBorder("Log"));
  reportScroll = new JScrollPane(report);
  p.add(reportScroll);
  resultAndLog.add(p);
  content.add(resultAndLog, "West");

  JMenuBar menuBar = new JMenuBar();

  setJMenuBar(menuBar);

  addConnectionMenu(menuBar);
  addTransactionMenu(menuBar);
  addTellMenu(menuBar);
  addUtilsMenu(menuBar);
  addAskMenu(menuBar);
  stateChanged();
  try {
    xmlConverter = new XMLToLisp();
  } catch (XMLConversionException e) {
    report (e.getMessage());
  }    
}  

  // Adds a menu allowing connection/disconnection to the server.
  
abstract void addConnectionMenu(JMenuBar menuBar);

  // Adds a menu controlling transactions.
void addTransactionMenu(JMenuBar menuBar) {
  transMenu = new JMenu("Transaction");
  transMenu.setMnemonic('t');

  JMenuItem item = new JMenuItem("Begin");
    item.setMnemonic('b');
  item.addActionListener(this);
  transMenu.add(item);

  item = new JMenuItem("Commit");
    item.setMnemonic('e');
  item.addActionListener(this);
  transMenu.add(item);

  item = new JMenuItem("Reset");
    item.setMnemonic('r');
  item.addActionListener(this);
  transMenu.add(item);

  item = new JMenuItem("Abort");
    item.setMnemonic('a');
  item.addActionListener(this);
  transMenu.add(item);

  menuBar.add(transMenu);
}

  // Adds a tell menu 
abstract void addTellMenu(JMenuBar menuBar);

  // Adds a utilities menu 
abstract void addUtilsMenu(JMenuBar menuBar);

  // Adds an ask menu
void addAskMenu(JMenuBar menuBar) {
  askMenu = new JMenu("Ask");
  askMenu.setMnemonic('a');

  JMenuItem item = new JMenuItem("Classifier Tells");
  item.setMnemonic('t');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Direct Subs");
  item.setMnemonic('b');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Direct Supers");
  item.setMnemonic('p');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("All Subs");
  item.setMnemonic('b');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("All Supers");
  item.setMnemonic('p');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Graph");
  item.setMnemonic('g');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Satisfiability");
  item.setMnemonic('s');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Subsumption");
  item.setMnemonic('u');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Equivalence");
  item.setMnemonic('e');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Classify");
  item.setMnemonic('c');
  item.addActionListener(this);
  askMenu.add(item);

  item = new JMenuItem("Classify and Graph");
  item.setMnemonic('l');
  item.addActionListener(this);
  askMenu.add(item);

  menuBar.add(askMenu);
}

  // The state of the client has changed -- update the fields accordingly. 
void stateChanged() {
  boolean c = connected();    
  transMenu.setEnabled(c);
  askMenu.setEnabled(c);
  if (c) {
    conn.setText("Connected");
    conn.setBackground(Color.red);
    conn.setForeground(Color.white);
  } else {
    conn.setText("");
    conn.setBackground(null);
    conn.setForeground(Color.black);
  }
  boolean t = transaction();
  tellMenu.setEnabled(t);
  if (t) {
    trans.setText("Transaction");
    trans.setBackground(Color.blue);
    trans.setForeground(Color.white);
  } else {
    trans.setText("");
    trans.setBackground(null);
    trans.setForeground(Color.black);
  }
  repaint();
}

  // Disable all the menus, and store a value which represents state
  // of the menus before disabling. 
  void disableMenus() {
  menuState = 0;
  if (connMenu.isEnabled()) {
    menuState = menuState | 1;
    connMenu.setEnabled(false);
  }
  if (transMenu.isEnabled()) {
    menuState = menuState | 2;
    transMenu.setEnabled(false);
  }
  if (tellMenu.isEnabled()) {
    menuState = menuState | 4;
    tellMenu.setEnabled(false);
  }
  if (askMenu.isEnabled()) {
    menuState = menuState | 8;
    askMenu.setEnabled(false);
  }
}

  // Enable menus based on the state that was remembered.
  void enableMenus() {
    connMenu.setEnabled((menuState&1)/1==1);
    transMenu.setEnabled((menuState&2)/2==1);
    tellMenu.setEnabled((menuState&4)/4==1);
    askMenu.setEnabled((menuState&8)/8==1);
  }
  
  // Controls the Serving status indicator
  void serving(boolean b) {
    if (b) {
      serv.setText("Serving");
      serv.setBackground(Color.green);
      serv.setForeground(Color.black);
    } else {
      serv.setText("");
      serv.setBackground(null);
      serv.setForeground(Color.black);
    }
    repaint();
  }
  
  // Controls the Compiling status indicator
  void compiling(boolean b) {
    if (b) {
      comp.setText("Compiling");
      comp.setBackground(Color.yellow);
      comp.setForeground(Color.black);
    } else {
      comp.setText("");
      comp.setBackground(null);
      comp.setForeground(Color.black);
    }
  }
  
public Dimension getPreferredSize()
  {
    return new Dimension(640,320);
  }
  
public void actionPerformed(ActionEvent e) {
  String cmd = e.getActionCommand();
  if (cmd.equals("Connect")) {
    this.connect();
  } 
  else if (cmd.equals("Disconnect")) {
    this.release();
  } 
  else if (cmd.equals("Begin")) {
    this.beginTransaction();
  } 
  else if (cmd.equals("Commit")) {
    this.endTransaction();
  } 
  else if (cmd.equals("Reset")) {
    this.resetTransaction();
  } 
  else if (cmd.equals("Abort")) {
    this.abortTransaction();
  } 
  else if (cmd.equals("Clear KB")) {
    this.clear();
  } 
  else if (cmd.equals("New Concept")) {
    this.addConcept();
  } 
  else if (cmd.equals("New Primitive")) {
    this.addPrimConcept();
  } 
//   else if (cmd.equals("Description")) {
//     this.showDescription();
//   } 
  else if (cmd.equals("All Supers")) {
    this.allSupers();
  } 
  else if (cmd.equals("All Subs")) {
    this.allSubs();
  } 
  else if (cmd.equals("Classifier Tells")) {
    this.showTells();
  } 
  else if (cmd.equals("Direct Supers")) {
    this.directSupers();
  } 
  else if (cmd.equals("Direct Subs")) {
    this.directSubs();
  } 
  else if (cmd.equals("Graph")) {
    this.showGraph();
  } 
  else if (cmd.equals("Satisfiability")) {
    this.checkSatisfiability();
  } 
  else if (cmd.equals("Subsumption")) {
    this.checkSubsumption();
  } 
  else if (cmd.equals("Equivalence")) {
    this.checkEquivalence();
  } 
  else if (cmd.equals("Classify")) {
    this.classify();
  } 
  else if (cmd.equals("Classify and Graph")) {
    this.classifyAndGraph();
  } 
  else {
    System.out.println(cmd);
  }
}

  /** Begin a transaction */
void beginTransaction() {
  if (connected()) {
    boolean res = ch.begin_transaction();
    updateTells();
    if (res) {
      stateChanged();
      report("Transaction Started");
      
    } else {
      report("!! Transaction Start Failed !!");
    }
  }
}

  /** End the current transaction */
void endTransaction() {
  int state;
  if (connected()) {
    WaitThread wt = new WaitThread() {
      public void performOperation() {
	serving(true);
	disableMenus();
	long s = System.currentTimeMillis();
	boolean res = ch.end_transaction();
	long t = System.currentTimeMillis();
	if (profile) {
	  System.out.print("Served in: ");
	  System.out.println(t-s);
	}  
	if (res) {
	  report("Transaction Committed");
	} else {
	  report("!! Transaction End Failed !!");
	  stateChanged();
	}
      }
      public void doneOperation() {
	/* stop serving */
	enableMenus();
	stateChanged();
	serving(false);
	updateTells();

      }
    };
    wt.start();
  }
}

  /** Reset the current transaction, throwing any facts away, but
      keeping the transaction open */
void resetTransaction() {
  if (connected()) {
    boolean res = ch.reset_transaction();
    updateTells();
    if (res) {
      report("Transaction Reset");
    } else {
      report("!! Transaction Reset Failed !!");
    }
  }
  stateChanged();
}

  /** Abort the transaction -- throw away all the facts and end the
      transaction */
void abortTransaction() {
  if (connected()) {
    boolean res = ch.abort_transaction();
    updateTells();
    if (res) {
      stateChanged();
      report("Transaction Aborted");
    } else {
      report("!! Transaction Abort Failed !!");
    }
  }
  stateChanged();
}
  
void updateTells() {
  tells.setText(ch.tells());
}

  /** Pop up a window showing the tells that the classifier has 
    been given. */
  void showTells() {
    class TellFrame extends JFrame implements ActionListener {
    private JTextArea jt;
      
      TellFrame(String ts) {
	super("Classifier Tells");
	jt = new JTextArea(40,80);
	JPanel p = new JPanel(new BorderLayout(5,5));
	jt.setFont(new Font("SanSerif", Font.PLAIN, 8));
	
	p.setBorder(new TitledBorder("Classifier Tells"));
	p.add(new JScrollPane(jt));
	
	jt.setText(ts);
	getContentPane().add(p, BorderLayout.CENTER);
	JMenuBar mb = new JMenuBar();
	setJMenuBar(mb);
	JMenu menu = new JMenu("File");
	menu.setMnemonic('f');
	JMenuItem item = new JMenuItem("Save");
	item.addActionListener(this);
	item.setMnemonic('s');
	menu.add(item);
	
	item = new JMenuItem("Quit");
	item.addActionListener(this);
	item.setMnemonic('q');
	menu.add(item);
	mb.add(menu);
      }
      
    public void actionPerformed(ActionEvent e) {
      String cmd = e.getActionCommand();
      if (cmd.equals("Quit")) {
	this.setVisible(false);
      } else if (cmd.equals("Save")) {
	this.save();
      }
    }
      
    public void save() {
      FileDialog fd = new FileDialog(this);
      fd.setFile("*.xml");
      fd.show();
      String fn = fd.getFile();
      if (fn!=null && !fn.equals("")) {
	try {
	  PrintStream out = new PrintStream(new FileOutputStream(fn));
	  out.println(jt.getText());
	} catch (FileNotFoundException ex) {
	  ex.printStackTrace();
	}
      }
    }
      
    public Dimension getPreferredSize()
      {
	return new Dimension(320,320);
      }
      
    } // TellFrame
    TellFrame tf = new TellFrame(ch.allTells());
    tf.setSize(tf.getPreferredSize());
    tf.setVisible(true);
  } // End of showtells

  /** TELL: Clear the knowledge base */
void clear() {
  if (connected()) {
    try {
      ch.clear();
      updateTells();
    } catch (InternalException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (TransactionRequiredException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    }
  }
}
    
  /** TELL: Launch a dialog and then add a concept. */
void addPrimConcept() {
  if (connected()) {
    String[] names = {"concept name: ", "description:  "};
    String[] defaults = {"",""};
    ConceptDefinitionForm df = new ConceptDefinitionForm(parentFrame(), "NEW" , "");
    df.show();
    if (df.aborted()) {
    } else {
      String nm = df.getName();
      try {
	String d = df.getDescription();
	String xmld = "";
	if (d!=null && !d.equals("")) {
	  Parser.ReInit(new StringReader(d));
	  xmld = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	}
	if (nm!=null) {
	  if (!nm.equals("")) {
	    try {
	      ch.defconcept(nm);
	      if (!xmld.equals("")) {
		ch.impliesC("<CONCEPT><PRIMITIVE NAME=\"" + 
			  nm + "\"/></CONCEPT>", xmld);
	      }
	      updateTells();
	    } catch (InternalException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (TransactionRequiredException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (OpUnimplementedException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (ExprErrorException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    }
	  }
	}
      } catch (ParseException ex) {
	report ("Parsing error");
	report (ex.getMessage());
      }
    }
  }
}

  /* TELL */
void addConcept() {
  if (connected()) {
    String[] names = {"concept name: ", "description:  "};
    String[] defaults = {"",""};
    ConceptDefinitionForm df = new ConceptDefinitionForm(parentFrame(), "NEW" , "");
    df.show();
    if (df.aborted()) {
    } else {
      String nm = df.getName();
      try {
	String d = df.getDescription();
	String xmld = "";
	if (d!=null && !d.equals("")) {
	  Parser.ReInit(new StringReader(d));
	  xmld = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	}
	if (nm!=null) {
	  if (!nm.equals("")) {
	    try {
	      ch.defconcept(nm);
	      if (!xmld.equals("")) {
		ch.equalC("<CONCEPT><PRIMITIVE NAME=\"" + 
			  nm + "\"/></CONCEPT>", xmld);
	      }
	      updateTells();
	    } catch (InternalException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (TransactionRequiredException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (OpUnimplementedException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    } catch (ExprErrorException ex) {
	      report (Short.toString(ex.code) + " " + ex.information);
	    }
	  }
	}
      } catch (ParseException ex) {
	report ("Parsing error");
	report (ex.getMessage());
      }
    }
  }
}

  /** Print the contents of the array to the result area */
  void resultArrays(String [][] arg) {
    result("(");
    int al = arg.length;
    for(int i=0; i < al;i++) {
      if (i>0) {
	result(" ");
      }
      result("(");
      int l = arg[i].length;
      for(int j=0; j < l; j++) {
	String str = arg[i][j];
	if (str!=null && !str.equals("")) {
	  if (j>0) {
	    result (" ");
	  }
	  result(convertToDescription(str));
	}
      }
      result(")");
    }
    result(")\n");
  }

  /** Display the contents of the array on the result */
  void resultArray(String [] arg) {
    result("(");
    for(int i=0; i < arg.length;i++) {
      if (arg[i]!=null && !arg[i].equals("")) {
	if (i>0) {
	  result(" ");
	}
	result(convertToDescription(arg[i]));
      }
    }
    result(")\n");
  }

  
public String convertToDescription(String xmlName) {
  String desc = "";
  try {
    desc = xmlConverter.convert(xmlName);
  } catch (XMLConversionException e) {
    System.err.println(xmlName);
    desc = e.getMessage();
  }
  return desc;
}
    
  /** ASK: Show the subs of a concept. */
  void directSubs() {
    clearResult();
    if (connected()) {
      String[][] res;
      String nm = getConceptNameFromUser(":TOP");
      if (nm!=null) {
	if (nm.equals(":TOP")) {
	  nm = "<TOP/>";
	} else if (nm.equals(":BOTTOM")) {
	  nm = "<BOTTOM/>";
	} else {
	  nm = "<PRIMITIVE NAME=\"" + nm + "\"/>";
	}
	try {
	  res = ch.directSubsC(nm);
	  resultArrays(res);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }

  /** ASK: Show the supers of a concept. */
  void directSupers() {
    clearResult();
    if (connected()) {
      String[][] res;
      String nm = getConceptNameFromUser(":TOP");
      if (nm!=null) {
	if (nm.equals(":TOP")) {
	  nm = "<TOP/>";
	} else if (nm.equals(":BOTTOM")) {
	  nm = "<BOTTOM/>";
	} else {
	  nm = "<PRIMITIVE NAME=\"" + nm + "\"/>";
	}
	try {
	  res = ch.directSupersC(nm);
	  resultArrays(res);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }
  /** ASK: Show all the subs of a concept. */
  void allSubs() {
    clearResult();
    if (connected()) {
      String[][] res;
      String nm = getConceptNameFromUser(":TOP");
      if (nm!=null) {
	if (nm.equals(":TOP")) {
	  nm = "<TOP/>";
	} else if (nm.equals(":BOTTOM")) {
	  nm = "<BOTTOM/>";
	} else {
	  nm = "<PRIMITIVE NAME=\"" + nm + "\"/>";
	}
	try {
	  res = ch.allSubsC(nm);
	  resultArrays(res);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }
  
  /** ASK: Show all the supers of a concept */
  void allSupers() {
    clearResult();
    if (connected()) {
      String[][] res;
      String nm = getConceptNameFromUser(":TOP");
      if (nm!=null) {
	if (nm.equals(":TOP")) {
	  nm = "<TOP/>";
	} else if (nm.equals(":BOTTOM")) {
	  nm = "<BOTTOM/>";
	} else {
	  nm = "<PRIMITIVE NAME=\"" + nm + "\"/>";
	}
	try {
	  res = ch.allSupersC(nm);
	  resultArrays(res);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
}
  
  /** Launch a dialog requesting some user input */
  String getConceptNameFromUser(String initial) {
    String[] names = {"concept name: "};
    String[] defaults = {initial};
    ArgumentForm af = new ArgumentForm(parentFrame(), 
				       names,
				       defaults);
    af.show();
    if (af.aborted()) {
      return null;
    }
    return af.getArg(0);
  }
  
  String getDescriptionFromUser(String initial) {
    String[] names = {"concept description: "};
    String[] defaults = {initial};
    ArgumentAreaForm df = new ArgumentAreaForm(parentFrame(), names, defaults);
    df.show();
    if (df.aborted()) {
      return null;
    }
    String s = df.getArg(0);
    String xmld = "";
    try {
      if (s!=null && !s.equals("")) {
	Parser.ReInit(new StringReader(s));
	xmld = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
      }
    } catch (ParseException ex) {
      report("Parsing error");
      report(ex.getMessage());
    }
    return xmld;
  }
  
  
  /** Open a window displaying a graph of a section of the model */
  void showGraph() {
    //  String nm = JOptionPane.showInputDialog(this, "Concept Name:");
    if (connected()) {
      String[] names = {"concept name: ", "depth: "};
      String[] defaults = {":TOP", ""};
      ArgumentForm af = new ArgumentForm(parentFrame(), 
					 names,
					 defaults);
      af.show();
      if (!af.aborted()) {
	String nm = af.getArg(0);
	String ds = af.getArg(1);
	int d = 0;
	if (ds!=null && ds.equals("")) {
	  d = 1000000;
	} else {
	  try {
	    d = Integer.parseInt(ds);
	  } catch (NumberFormatException ex) {
	    d = -1;
	  }
	}
	if (nm!=null && d>0) {
	  if (gr==null) {
	    gr = new Grapher(this, ch);
	  }
	  gr.showHierarchy(nm,d);
	}
      }
    }
  }
  
  void checkSatisfiability() {
    if (connected()) {
      String d = getDescriptionFromUser(":TOP");
      if (d!=null && !d.equals("")) {
	clearResult();
	try {
	  if (ch.satisfiable(d)) {
	    result("YES");
	  } else {
	    result("NO");
	  }	
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }
  
  void checkSubsumption() {
    if (connected()) {
    String[] names = {"description1: ", "description2: "};
    String[] defaults = {"", ""};
    ArgumentAreaForm af = new ArgumentAreaForm(parentFrame(), 
					       names,
					       defaults);
    af.show();
    if (!af.aborted()) {
      try {
	String d1 = af.getArg(0);
	String d2 = af.getArg(1);
	String xmld1 = "";
	String xmld2 = "";
	Parser.ReInit(new StringReader(d1));
	xmld1 = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	Parser.ReInit(new StringReader(d2));
	xmld2 = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	clearResult();
	try {
	  if (ch.subsumes(xmld1,xmld2)) {
	    result("YES");
	  } else {
	    result("NO");
	  }	
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      } catch (ParseException ex) {
	report ("Parsing error");
	report (ex.getMessage());
      }
    }
  }
}

  void checkEquivalence() {
    if (connected()) {
      String[] names = {"description1", "description2"};
      String[] defaults = {"", ""};
      ArgumentAreaForm af = new ArgumentAreaForm(parentFrame(), 
						 names,
						 defaults);
      af.show();
      if (!af.aborted()) {
	try {
	  String d1 = af.getArg(0);
	  String d2 = af.getArg(1);
	  String xmld1 = "";
	  String xmld2 = "";
	  Parser.ReInit(new StringReader(d1));
	  xmld1 = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	  Parser.ReInit(new StringReader(d2));
	  xmld2 = "<CONCEPT> " + Parser.description() + "</CONCEPT>";
	  
	  clearResult();
	  try {
	    if (ch.equivalent(xmld1,xmld2)) {
	      result("YES");
	    } else {
	      result("NO");
	    }	
	  } catch (InternalException ex) {
	    report (Short.toString(ex.code) + " " + ex.information);
	  } catch (KBModifiedException ex) {
	    report (Short.toString(ex.code) + " " + ex.information);
	  } catch (OpUnimplementedException ex) {
	    report (Short.toString(ex.code) + " " + ex.information);
	  } catch (ExprErrorException ex) {
	    report (Short.toString(ex.code) + " " + ex.information);
	  }
	} catch (ParseException ex) {
	  report ("Parsing error");
	  report (ex.getMessage());
	}
      }
    }
  }
  
  void classify() {
    clearResult();
    if (connected()) {
      String d = getDescriptionFromUser(":TOP");
      if (d!=null && !d.equals("")) {
	try {
	  taxonomyNode tN =  (ch.taxonomyPosition(d));
	  resultArrays(tN.supers);
	  resultArrays(tN.subs);
	  resultArray(tN.equivs);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }
  
  void classifyAndGraph() {
    clearResult();
    if (connected()) {
      String d = getDescriptionFromUser(":TOP");
      if (d!=null && !d.equals("")) {
	try {
	  taxonomyNode tN =  (ch.taxonomyPosition(d));
	  resultArrays(tN.supers);
	  resultArrays(tN.subs);
	  resultArray(tN.equivs);
	  if (gr==null) {
	    gr = new Grapher(this, ch);
	  }
	  gr.showHierarchy(d,tN, 100);
	} catch (InternalException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (KBModifiedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (OpUnimplementedException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	} catch (ExprErrorException ex) {
	  report (Short.toString(ex.code) + " " + ex.information);
	}
      }
    }
  }
  
  /** Show some results */
  void result(String str) {
    result.append(str);
  }
  
  /** Clear the results display */
  void clearResult() {
    result.setText("");
  }
  
  /** Report something */
  void report(String str) {
    report.append(str + "\n");
    JScrollBar sb = reportScroll.getVerticalScrollBar();
    sb.setValue(sb.getMaximum());
  }
  
  /** Report something */
  void reportNoNL(String str) {
    report.append(str);
  }
  
  
  /** Try and connect to the server */
public void connect() { 
  if (!connected()) {
    report("Connecting");
    if (cc==null) {
      cc = new ClientConnector(defaultNSHost,
			       defaultNSPort, 
			       defaultServerName,
			       "client");
    }
    ch = cc.connect(parentFrame());
    if (connected()) {
      // setTitle("Client -- " + ch.identifier());
      stateChanged();
    }
  }
}
  /** Explicitly use this client handler */
public void connect(ClientHandler handler) {
  ch = handler;
  stateChanged();
}
  
  /** Release the client handler. If it's in a transaction, the
    transaction will be aborted */
  void release() {
    if (connected()) {
      if (ch.in_transaction()) {
	report ("Aborting transaction...");
	ch.abort_transaction();
      }
      report("Disconnecting");
      try {
	ch.release();
      } catch (Exception e) {
	System.err.println(e.toString());
      };
    }
    ch = null;
    stateChanged();
  }
  
  /** Is the client connected to the server? */
  boolean connected() {
    return ch!=null;
  }
  
  /** Are we in a transaction? */
  boolean transaction() {
    return (connected() && ch.in_transaction());
  }
  
  /** Operation needed for Compiler Logger */
public void compilerMessage(String msg) {
  report(msg);
}
  
  JFrame parentFrame() {
    return null;
  }
  
  //private JCheckBox trans;
  JMenuBar menuBar;
  JMenu connMenu, transMenu, tellMenu, askMenu;
  
  JTextField trans, conn, comp, serv;
  JTextArea tells;
  JTextArea report, result;
  JScrollPane reportScroll, resultScroll;
  JButton clear;
  ClientHandler ch;
  Classifier cl;
  Grapher gr;
  ClientConnector cc;
  int menuState = 0;
  
  String defaultNSHost;
  
public void setDefaultNSHost(String h) {
  defaultNSHost = h;
}
  
  String defaultNSPort;
  
public void setDefaultNSPort(String h) {
  defaultNSPort = h;
}
  
  String defaultServerName;
  
public void setDefaultServerName(String h) {
  defaultServerName = h;
}
  
  boolean profile = false;
}

