/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER
/// and THE UNIVERSITY OF MANCHESTER, seanb@cs.man.ac.uk

////////////////////////////////////////////////////////////////////////////
/// FaCT Client                                                          ///
/// COPYRIGHT (C) 1997, 1998, 1999 SEAN K. BECHHOFER                     ///
/// and THE UNIVERSITY OF MANCHESTER                                     /// 
///					                                 ///
/// This program is free software; you can redistribute it and/or        ///
/// modify it under the terms of the GNU General Public License          ///
/// as published by the Free Software Foundation; either version 2       ///
/// of the License, or (at your option) any later version.               ///
///                                                                      ///
/// This program is distributed in the hope that it will be useful,      ///
/// but WITHOUT ANY WARRANTY; without even the implied warranty of       ///
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        ///
/// GNU General Public License for more details.                         ///
///                                                                      ///
/// You should have received a copy of the GNU General Public License    ///
/// along with this program; if not, write to the Free Software          ///
/// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            ///
/// 02111-1307, USA.                                                     ///
///                                                                      ///
/// Enquiries about FaCT should be directed to:                          ///
///                                                                      ///
/// email: horrocks@cs.man.ac.uk                                         ///
/// www:   http://www.cs.man.ac.uk/~horrocks                             ///
/// smail: Ian Horrocks, Department of Computer Science, Oxford Road,    ///
///       Manchester 467M13 9PL, United Kingdom                          ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
///                                                                      ///
/// This class is intended to demonstrate how to write a Java client     ///
/// that will communicate with the FaCT description logic reasoner via   ///
/// the CORBA-based interface. It is not supported or maintained and no  ///
/// guarantees are supplied as to the quality or robustness of the code. ///
/// SKB, 4/10/99                                                         ///
///                                                                      ///
////////////////////////////////////////////////////////////////////////////

package img.fact.client;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.applet.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.net.*;
import lispwrapper.*;
import org.omg.CORBA.*;
import org.omg.CosNaming.*;
import img.fact.cexpr.*;
import img.fact.*;
import img.util.*;

/** Application version of the FaCT client.
 * @author SKB */

public class ClientApplication extends BasicClient {

JFrame parentFrame;
JFrame parentFrame() {
  return parentFrame;
}

public ClientApplication(JFrame parent) {
  profile = (System.getProperty("profile")!=null);

  parentFrame = parent;
  parentFrame.setTitle("FaCT Client");
}

public void actionPerformed(ActionEvent e) {
  // Connect; Disconnect; Quit; Begin; End; Reset; Abort
  String cmd = e.getActionCommand();
  if (cmd.equals("Quit")) {
    this.closeWindowAndExit();
  } else if (cmd.equals("Compile File")) {
    this.compile();
  } else if (cmd.equals("Convert File")) {
    this.convert();
  } else {
    super.actionPerformed(e);
  }
}

void addConnectionMenu(JMenuBar menuBar) {
  connMenu = new JMenu("Connection");
  connMenu.setMnemonic('c');

  JMenuItem item = new JMenuItem("Connect");
  item.setMnemonic('c');
  item.addActionListener(this);
  connMenu.add(item);

  item = new JMenuItem("Disconnect");
  item.setMnemonic('d');
  item.addActionListener(this);
  connMenu.add(item);

  item = new JMenuItem("Quit");
  item.setMnemonic('q');
  item.addActionListener(this);
  connMenu.add(item);

  menuBar.add(connMenu);
}

void addTellMenu(JMenuBar menuBar) {
  tellMenu = new JMenu("Tell");
  tellMenu.setMnemonic('t');

  JMenuItem item = new JMenuItem("Clear KB");
  item.setMnemonic('c');
  item.addActionListener(this);
  tellMenu.add(item);

  item = new JMenuItem("New Concept");
  item.setMnemonic('n');
  item.addActionListener(this);
  tellMenu.add(item);

  item = new JMenuItem("New Primitive");
  item.setMnemonic('p');
  item.addActionListener(this);
  tellMenu.add(item);

  item = new JMenuItem("Compile File");
  item.setMnemonic('o');
  item.addActionListener(this);
  tellMenu.add(item);

  menuBar.add(tellMenu);
}

void addUtilsMenu(JMenuBar menuBar) {
  JMenu utilMenu = new JMenu("Utilities");
  utilMenu.setMnemonic('u');
  
  JMenuItem item = new JMenuItem("Convert File");
  item.setMnemonic('c');
  item.addActionListener(this);
  utilMenu.add(item);

  menuBar.add(utilMenu);
}

  /** A class which governs compilation of knowledge bases. */
  class CompilerThread extends Thread {
    boolean notDone = true;
    BasicClient lc;
    String file;
    
    CompilerThread(BasicClient c, String f) {
      lc = c;
	file = f;
    }
    
  public void run() {
    lc.compiling(true);
    lc.disableMenus();
    try {
      Compiler c = new Compiler(ch);
      c.setLogger(lc);
      c.clearBeforeCompiling(false);
      long s = System.currentTimeMillis();
      
      c.compile(new File(file));
      
      lc.updateTells();
      long t = System.currentTimeMillis();
      if (profile) {
	System.out.print("Compiled in: ");
	System.out.println(t-s);
      }  
    } catch (Compiler.CompilerException ex) {
      lc.report("Compilation Problem: " + ex.information);
    }
    lc.enableMenus();
    lc.compiling(false);
  }
  }

  /* TELL */
private void compile() {
  FileDialog fd = new FileDialog(parentFrame());
  fd.setFile("*.xml");
  fd.show();
  if (fd.getFile()!=null) {
    String fn = fd.getDirectory() + fd.getFile();
    CompilerThread ct = new CompilerThread(this, fn);
    ct.start();
  }
}

private void convert() {
  FileDialog fd = new FileDialog(parentFrame());
  fd.setFile("*.kb");
  fd.show();
  if (fd.getFile()!=null) {
    String in = fd.getDirectory() + fd.getFile();
    FileDialog ofd = new FileDialog(parentFrame());
    ofd.setDirectory(fd.getDirectory());
    ofd.setMode(FileDialog.SAVE);
    String tmp = fd.getFile();
    int i = tmp.lastIndexOf('.');
    i = i < 0 ? tmp.length()-1 : i;
    ofd.setFile(tmp.substring(0, i) + ".xml");
    ofd.show();
    if (ofd.getFile()!=null) {
      try {
	String out = ofd.getDirectory() + ofd.getFile();
	img.fact.lisp2xml.Parser parser = 
	  new img.fact.lisp2xml.Parser(new FileReader(new File(in)));
	img.fact.lisp2xml.Parser.setImpliedDefs(true);
	String r = parser.kb();
	PrintStream op = new PrintStream(new FileOutputStream(new File(out)));
	op.println(r);
      } catch (img.fact.lisp2xml.ParseException x) {
	report("Conversion failed!");
	report(x.getMessage());
      } catch (java.io.IOException x) {
	report("Conversion failed!");
	report(x.getMessage());
      }
      report("Conversion succeeded ok");
    }
  }
}

  /** Do any clearing up required, disconnecting etc. */
public void closeWindowAndExit() {
  release();
  System.exit(0);
}
}







