package img.fact.client;

import java.util.*;
import java.io.*;
import java.net.*;
import lispwrapper.*;
import img.fact.sexpr.*;
import img.fact.xml.*;
import img.fact.*;

import org.w3c.dom.*;
import org.xml.sax.*;
import org.xml.sax.helpers.ParserFactory;

import com.sun.xml.tree.*;
import com.sun.xml.parser.Resolver;

/** A simple little class that produces a graph of a FaCT
 * model that can be read by graphing applications
 */
public class DotGrapher {

private Classifier cl;
private ClientHandler ch;

private XMLToName xtn;

public DotGrapher() { 
  String ior;
  String ref=null;
  ClientConnector cc = new ClientConnector("localhost",
					   "8000",
					   "FaCTServer",
					   "grapher");
  ch = cc.connect();
  initializeXML();
}

public DotGrapher(String host, String port, String serverName, String name) { 
  String ior;
  String ref=null;
  ClientConnector cc = new ClientConnector(host,
					   port,
					   serverName,
					   name);
  ch = cc.connect();
  initializeXML();
}

public DotGrapher(ClientHandler c) {
  ch = c;
  initializeXML();
}

public void initializeXML() {
  try {
    xtn = new XMLToName();
  } catch (XMLConversionException e) {
    System.err.println("Error: " + e.getMessage());
    e.printStackTrace();
    System.exit(99);
  }
}

public void release() {
  if (ch!=null) {
    ch.release();
  }
}

public void graph(String top, PrintStream os) throws XMLConversionException {
  printHeader(os);
  if (top.equals(":TOP")) {
    top = "<TOP/>";
  } else {
    top = "<PRIMITIVE NAME=\"" + top + "\"/>";
  }
  graphConcept(top, os);
  try {
    String [][] res = ch.allSubsC(top);
    for (int i=0; i<res.length; i++) {
      graphConcept(res[i][0],os);
    }
    printFooter(os);
  } catch (InternalException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (KBModifiedException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (OpUnimplementedException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  } catch (ExprErrorException ex) {
    report (Short.toString(ex.code) + " " + ex.information);
  }
}

private void printHeader(PrintStream os) {
  os.println("digraph generated {");
  //  os.println("  size=\"7,10\";");
  os.println("  ratio=fill;");
  os.println("  node [style=filled];");
  os.println("  node [color=red];");
  os.println("  node [fontcolor=white];");
}

private void printFooter(PrintStream os) {
  os.println("}");
}

private void graphConcept(String con, PrintStream os) throws XMLConversionException {
  if (!con.equals("<BOTTOM/>")) {
    String name = "";
    try {
      name = xtn.convert(con);
    } catch (XMLConversionException e) {
      System.err.println(con);
      throw e;
    }
    os.println();
    if (name.equals(":TOP")) {
      name = "TOP";
    };
    os.println("  " + name.toLowerCase() + ";");
    os.print("  " + name.toLowerCase() + "-> {");
    try {
      // Find all the subs, and add a link
      String [][] res = ch.directSubsC(con);
      for (int i=0; i<res.length; i++) {
	String bt = xtn.convert(res[i][0]);
	if (!bt.equals(":BOTTOM")) {
	  os.print(bt.toLowerCase() + "; ");
	}
      }
      os.println("};");
    } catch (InternalException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (KBModifiedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (OpUnimplementedException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    } catch (ExprErrorException ex) {
      report (Short.toString(ex.code) + " " + ex.information);
    }
  }
}

private void report(String msg) {
  System.err.println(msg);
}

public static void main(String[] args) {

  Hashtable argSpec = new Hashtable();
  argSpec.put("nsHost", new Integer(1));
  argSpec.put("nsPort", new Integer(1));
  argSpec.put("serverName", new Integer(1));
  argSpec.put("top", new Integer(1));
  argSpec.put("help", new Integer(0));
  
  img.util.CommandLineArguments cla = 
    new img.util.CommandLineArguments(argSpec);

  // Extract the arguments
  String serverName = "FaCTServer";
  String nsHost = "localhost";
  String nsPort = "8000";
  String fileName = "";
  String top = ":TOP";
  try {
    String[] tmp;
    cla.setArgs(args);
    if (cla.flagPresent("help")) {
      printHelp();
      System.exit(0);
    }
    tmp = cla.argsForSelector("serverName");
    if (tmp!=null) {
      serverName = tmp[0];
    }
    tmp = cla.argsForSelector("nsHost");
    if (tmp!=null) {
      nsHost = tmp[0];
    }
    tmp = cla.argsForSelector("nsPort");
    if (tmp!=null) {
      nsPort = tmp[0];
    }
    tmp = cla.argsForSelector("top");
    if (tmp!=null) {
      top = tmp[0];
    } else {
      top = ":TOP";
    }
    tmp = cla.otherArgs();
    fileName = tmp[0];
  } catch (Exception ex) {
    System.err.println("Usage: java img.fact.client.DotGrapher [-help] [-nsHost host] [-nsPort port] [-serverName name] [-top name] file");
    System.exit(1);
  }

  DotGrapher dg = new DotGrapher(nsHost, nsPort, serverName, "grapher");
  try {
    PrintStream os = new PrintStream(new FileOutputStream(fileName));
    dg.graph(top, os);
    dg.release();
  } catch (Exception ex) {
    ex.printStackTrace();
    dg.release();
    System.err.println("ERROR: " + ex.toString());
    System.exit(1);
  }
}

public static void printHelp() {
  System.out.println("Usage: java img.fact.client [options] file");
  System.out.println("Produces a static dump of the hierarchy as a graph");
  System.out.println("\t-nsHost\t\tHost running naming service");
  System.out.println("\t-nsPort\t\tPort running naming service");
  System.out.println("\t-serverName\t\tName the server is bound to");
  System.out.println("\t-top\t\tTop concept to start from");
  System.out.println("\t-help\t\tPrint this message");
}

}
